/* $Header: ntp_config.c,v 1.3 91/06/19 11:06:51 ath Exp $ */
/*
 * ntp_config.c - read and apply configuration information
 */
#include <stdio.h>
#include <ctype.h>

#include "global.h"
#include "socket.h"
#include "sockaddr.h"

#include "ntp_types.h"
#include "ntp_syslog.h"
#include "ntp_fp.h"
#include "ntp.h"

/*
 * These routines are used to read configuration information from 
 * command lines given to the NOS "ntp" command.  To configure
 * ntp, you include multiple "ntp" commands in the NOS startup
 * file, each with a line from the desired configuration.

/*
 * We understand the following configuration entries.
 *
 * peer 128.100.1.1 [ version 2 ] [ key 5 ] [ minpoll ]
 * server 128.100.2.2 [ version 1 ] [ key 6 ] [ minpoll ]
 * precision -7
 * broadcast 128.100.224.255 [ version 2 ] [ key 2 ] [ minpoll ]
 * broadcastclient yes|no
 * broadcastdelay 0.0102
 * authenticate yes|no
 * monitor yes|no
 * authdelay 0.00842
 * restrict 128.100.100.0 [ mask 255.255.255.0 ] ignore|noserve|notrust|noquery
 * driftfile file_name
 * keys file_name
 * debug off|dec|inc
 * And then some.  See the manual page.
 */

/*
 * Types of entries we understand.
 */
#define CONFIG_UNKNOWN		0

#define	CONFIG_PEER		1
#define	CONFIG_SERVER		2
#define	CONFIG_PRECISION	3
#define	CONFIG_SETDRIFT		4
#define	CONFIG_BROADCAST	5
#define	CONFIG_BROADCASTCLIENT	6
#define	CONFIG_AUTHENTICATE	7
#define	CONFIG_KEYS		8
#define	CONFIG_MONITOR		9
#define	CONFIG_AUTHDELAY	10
#define	CONFIG_RESTRICT		11
#define	CONFIG_BDELAY		12
#define	CONFIG_TRUSTEDKEY	13
#define	CONFIG_REQUESTKEY	14
#define	CONFIG_CONTROLKEY	15
#define	CONFIG_TRAP		16
#define	CONFIG_FUDGE		17
#define	CONFIG_MAXSKEW		18
#define	CONFIG_RESOLVER		19
#define	CONFIG_SELECT		20
#define CONFIG_DEBUG		21

#define	CONF_MOD_VERSION	1
#define	CONF_MOD_KEY		2
#define	CONF_MOD_MINPOLL	3

#define	CONF_RES_MASK		1
#define	CONF_RES_IGNORE		2
#define	CONF_RES_NOSERVE	3
#define	CONF_RES_NOTRUST	4
#define	CONF_RES_NOQUERY	5
#define	CONF_RES_NOMODIFY	6
#define	CONF_RES_NOPEER		7
#define	CONF_RES_NOTRAP		8
#define	CONF_RES_LPTRAP		9
#define	CONF_RES_NTPPORT	10

#define	CONF_TRAP_PORT		1
#define	CONF_TRAP_INTERFACE	2

#define	CONF_FDG_TIME1		1
#define	CONF_FDG_TIME2		2
#define	CONF_FDG_VALUE1		3
#define	CONF_FDG_VALUE2		4
#define	CONF_FDG_FLAG1		5
#define	CONF_FDG_FLAG2		6
#define	CONF_FDG_FLAG3		7
#define	CONF_FDG_FLAG4		8

#define CONF_DBG_OFF		1
#define CONF_DBG_INC		2
#define CONF_DBG_DEC		3

/*
 * Translation table - keywords to function index
 */
struct keyword {
	char *text;
	int keytype;
};

struct keyword keywords[] = {
	{ "peer",		CONFIG_PEER },
	{ "server",		CONFIG_SERVER },
	{ "precision",		CONFIG_PRECISION },
	{ "drift",		CONFIG_SETDRIFT },
	{ "broadcast",		CONFIG_BROADCAST },
	{ "broadcastclient",	CONFIG_BROADCASTCLIENT },
#ifdef XNTP_AUTHENTICATE
	{ "authenticate",	CONFIG_AUTHENTICATE },
	{ "keys",		CONFIG_KEYS },
#endif /* XNTP_AUTHENTICATE */
#ifdef XNTP_MONITOR
	{ "monitor",		CONFIG_MONITOR },
#endif  /* XNTP_MONITOR */
#ifdef XNTP_AUTHENTICATE
	{ "authdelay",		CONFIG_AUTHDELAY },
#endif  /* XNTP_AUTHENTICATE */
	{ "restrict",		CONFIG_RESTRICT },
	{ "broadcastdelay",	CONFIG_BDELAY },
#ifdef XNTP_AUTHENTICATE
	{ "trustedkey",		CONFIG_TRUSTEDKEY },
	{ "requestkey",		CONFIG_REQUESTKEY },
	{ "controlkey",		CONFIG_CONTROLKEY },
#endif  /* XNTP_AUTHENTICATE */
#ifdef XNTP_MODE6
	{ "trap",		CONFIG_TRAP },
#endif  /* XNTP_MODE6 */
#ifdef REFCLOCK
	{ "fudge",		CONFIG_FUDGE },
#endif /* REFCLOCK */
	{ "maxskew",		CONFIG_MAXSKEW },
/*	{ "resolver",		CONFIG_RESOLVER },*/
	{ "select",		CONFIG_SELECT },
#ifdef DEBUG
	{ "debug",		CONFIG_DEBUG },
#endif  /* DEBUG */
	{ "",			CONFIG_UNKNOWN }
};

/*
 * Modifier keywords
 */
struct keyword mod_keywords[] = {
	{ "version",	CONF_MOD_VERSION },
	{ "key",	CONF_MOD_KEY },
	{ "minpoll",	CONF_MOD_MINPOLL },
	{ "",		CONFIG_UNKNOWN }
};


/*
 * Special restrict keywords
 */
struct keyword res_keywords[] = {
	{ "mask",	CONF_RES_MASK },
	{ "ignore",	CONF_RES_IGNORE },
	{ "noserve",	CONF_RES_NOSERVE },
	{ "notrust",	CONF_RES_NOTRUST },
	{ "noquery",	CONF_RES_NOQUERY },
	{ "nomodify",	CONF_RES_NOMODIFY },
	{ "nopeer",	CONF_RES_NOPEER },
	{ "notrap",	CONF_RES_NOTRAP },
	{ "lowpriotrap",	CONF_RES_LPTRAP },
	{ "ntpport",	CONF_RES_NTPPORT },
	{ "",		CONFIG_UNKNOWN }
};


/*
 * Keywords for the trap command
 */
struct keyword trap_keywords[] = {
	{ "port",	CONF_TRAP_PORT },
	{ "interface",	CONF_TRAP_INTERFACE },
	{ "",		CONFIG_UNKNOWN }
};


/*
 * Keywords for the fudge command
 */
#ifdef REFCLOCK
struct keyword fudge_keywords[] = {
	{ "time1",	CONF_FDG_TIME1 },
	{ "time2",	CONF_FDG_TIME2 },
	{ "value1",	CONF_FDG_VALUE1 },
	{ "value2",	CONF_FDG_VALUE2 },
	{ "flag1",	CONF_FDG_FLAG1 },
	{ "flag2",	CONF_FDG_FLAG2 },
	{ "flag3",	CONF_FDG_FLAG3 },
	{ "flag4",	CONF_FDG_FLAG4 },
	{ "",		CONFIG_UNKNOWN }
};
#endif /* REFCLOCK */

/*
 * Keywords for the debug command 
 */
struct keyword debug_keywords[] = {
	{ "off",	CONF_DBG_OFF },
	{ "inc",	CONF_DBG_INC },
	{ "dec",	CONF_DBG_DEC },
	{ "",		CONFIG_UNKNOWN }
};


/*
 * Miscellaneous macros
 */
/*#define	STREQ(a, b)		(*(a) == *(b) && strcmp((a), (b)) == 0)*/

/*
 * Definitions of things either imported from or exported to outside
 */
#ifdef DEBUG
extern int debug;
#endif

static int matchkey();
static int getnetnum();

/*
 * getconfig - interpret a single command line (already parsed into words
 * by NOS).
 */
void
getconfig(argc, argv)
	int argc;
	char *argv[];
{
	register int i;
	int c;
	int errflg;
	int peerversion;
	u_long peerkey;
	int peerflags;
	int hmode;
	struct sockaddr_in peeraddr;
	struct sockaddr_in maskaddr;
	char **tokens;
	int ntokens;
	int tok;
	struct interface *localaddr;
	char *config_file;
#ifdef REFCLOCK
	struct refclockstat clock;
#endif /* REFCLOCK */

	extern void proto_config();
#ifdef XNTP_AUTHENTICATE
	extern void authtrust();
	extern int authreadkeys();
	extern u_long info_auth_keyid;
#endif  /* XNTP_AUTHENTICATE */
	extern struct peer *peer_config();
	extern void stats_config();
#ifdef XNTP_MONITOR
	extern void mon_start(), mon_stop();
#endif  /* XNTP_MONITOR */
	extern int atolfp();
	extern struct interface *findinterface();
	extern char *ntoa();

	/*
	 * Initialize, initialize
	 */
	errflg = 0;

	ntokens = argc - 1;
	tokens = argv + 1;	/* Rename argv. */
	if ( (tok = matchkey (tokens[0], keywords)) != CONFIG_UNKNOWN)
	  {
		switch(tok) {
		case CONFIG_PEER:
		case CONFIG_SERVER:
		case CONFIG_BROADCAST:
			if (tok == CONFIG_PEER)
				hmode = MODE_ACTIVE;
			else if (tok == CONFIG_SERVER)
				hmode = MODE_CLIENT;
			else
				hmode = MODE_BROADCAST;

			if (ntokens < 2) {
				tprintf("No address for %s, line ignored\n",
					tokens[0]);
				break;
			}

			if (!getnetnum(tokens[1], &peeraddr, 0)) {
				tprintf("IP address required\n");
			} else {
				errflg = 0;

				if (ISBADADR(&peeraddr)) {
					tprintf(
				    "attempt to configure invalid address %s\n",
					    ntoa(&peeraddr));
					break;
				}
			}
			
			peerversion = NTP_VERSION;
			peerkey = 0;
			peerflags = 0;
			for (i = 2; i < ntokens; i++)
				switch (matchkey(tokens[i], mod_keywords)) {
				case CONF_MOD_VERSION:
					if (i >= ntokens-1) {
						tprintf(
				"peer/server version requires an argument\n");
						errflg = 1;
						break;
					}
					peerversion = atoi(tokens[++i]);
					if (peerversion != NTP_VERSION
					    && peerversion != NTP_OLDVERSION) {
						tprintf(
			     "inappropriate version number %s, line ignored\n",
						    tokens[i]);
						errflg = 1;
					}
					break;

				case CONF_MOD_KEY:
					/*
					 * XXX
					 * This is bad because atoi
					 * returns 0 on errors.  Do
					 * something later.
					 */
					if (i >= ntokens-1) {
						tprintf(
				      "peer/server key requires an argument\n");
						errflg = 1;
						break;
					}
					peerkey = (u_long)atoi(tokens[++i]);
					peerflags |= FLAG_AUTHENABLE;
					break;

				case CONF_MOD_MINPOLL:
					peerflags |= FLAG_MINPOLL;
					break;

				case CONFIG_UNKNOWN:
					errflg = 1;
					break;
				}
			if (errflg == 0) {
				if (peer_config(&peeraddr,
				    (struct interface *)0,
				    hmode, peerversion, peerkey,
				    peerflags) == 0) {
					tprintf(
					    "configuration of %s failed\n",
					    ntoa(&peeraddr));
				}
			}
			break;

		case CONFIG_PRECISION:
			if (ntokens >= 2) {
				i = atoi(tokens[1]);
				if (i >= 0 || i < -25)
					tprintf(
				"unlikely precision %s, line ignored\n",
					    tokens[1]);
				else
					proto_config(PROTO_PRECISION, i);
			}
			break;

		case CONFIG_SETDRIFT:
			if (ntokens >= 2)
				stats_config(STATS_FREQ, tokens[1]);
			else
			  {
			    extern l_fp drift_comp;
			    extern char *lfptoa();
			    tprintf ("Current drift value %s\n",
				     lfptoa(&drift_comp, 9));
			  }
			break;
		
		case CONFIG_BROADCASTCLIENT:
			errflg = 0;
			if (ntokens >= 2) {
				if (STREQ(tokens[1], "yes"))
					proto_config(PROTO_BROADCLIENT, 1);
				else if (STREQ(tokens[1], "no"))
					proto_config(PROTO_BROADCLIENT, 0);
				else
					errflg++;
			} else {
				errflg++;
			}

			if (errflg)
				tprintf(
				    "should be `broadcastclient yes|no'\n");
			break;
		
#ifdef XNTP_AUTHENTICATE
		case CONFIG_AUTHENTICATE:
			errflg = 0;
			if (ntokens >= 2) {
				if (STREQ(tokens[1], "yes"))
					proto_config(PROTO_AUTHENTICATE, 1);
				else if (STREQ(tokens[1], "no"))
					proto_config(PROTO_AUTHENTICATE, 0);
				else
					errflg++;
			} else {
				errflg++;
			}

			if (errflg)
				tprintf(
				    "should be `authenticate yes|no'\n");
			break;
#endif  /* XNTP_AUTHENTICATE */

#ifdef XNTP_MONITOR
		case CONFIG_MONITOR:
			errflg = 0;
			if (ntokens >= 2) {
				if (STREQ(tokens[1], "yes"))
					mon_start();
				else if (STREQ(tokens[1], "no"))
					mon_stop();
				else
					errflg++;
			} else {
				errflg++;
			}

			if (errflg)
				tprintf(
				    "should be `monitor yes|no'\n");
			break;
#endif  /* XNTP_MONITOR */
		
#ifdef XNTP_AUTHENTICATE
		case CONFIG_AUTHDELAY:
			if (ntokens >= 2) {
				l_fp tmp;

				if (!atolfp(tokens[1], &tmp)) {
					tprintf(
					    "authdelay value %s undecodable\n",
					    tokens[1]);
				} else if (tmp.l_ui != 0) {
					tprintf(
					    "authdelay value %s is unlikely\n",
					    tokens[1]);
				} else {
					proto_config(PROTO_AUTHDELAY, tmp.l_f);
				}
			}
			break;
#endif  /* XNTP_AUTHENTICATE */
		
		case CONFIG_RESTRICT:
			if (ntokens < 2) {
				tprintf( "restrict requires an address\n");
				break;
			}
			if (STREQ(tokens[1], "default"))
				peeraddr.sin_addr.s_addr = INADDR_ANY;
			else if (!getnetnum(tokens[1], &peeraddr, 1))
				break;
			
			/*
			 * Use peerversion as flags, peerkey as mflags.  Ick.
			 */
			peerversion = 0;
			peerkey = 0;
			errflg = 0;
			maskaddr.sin_addr.s_addr = ~0;
			for (i = 2; i < ntokens; i++) {
				switch (matchkey(tokens[i], res_keywords)) {
				case CONF_RES_MASK:
					if (i >= ntokens-1) {
						tprintf(
					       "mask keyword needs argument\n");
						errflg++;
						break;
					}
					i++;
					if (!getnetnum(tokens[i], &maskaddr, 1))
						errflg++;
					break;

				case CONF_RES_IGNORE:
					peerversion |= RES_IGNORE;
					break;
				
				case CONF_RES_NOSERVE:
					peerversion |= RES_DONTSERVE;
					break;
				
				case CONF_RES_NOTRUST:
					peerversion |= RES_DONTTRUST;
					break;
				
				case CONF_RES_NOQUERY:
					peerversion |= RES_NOQUERY;
					break;

				case CONF_RES_NOMODIFY:
					peerversion |= RES_NOMODIFY;
					break;

				case CONF_RES_NOPEER:
					peerversion |= RES_NOPEER;
					break;

				case CONF_RES_NOTRAP:
					peerversion |= RES_NOTRAP;
					break;

				case CONF_RES_LPTRAP:
					peerversion |= RES_LPTRAP;
					break;

				case CONF_RES_NTPPORT:
					peerkey |= RESM_NTPONLY;
					break;

				case CONFIG_UNKNOWN:
					errflg++;
					break;
				}
			}
			if (SRCADR(&peeraddr) == INADDR_ANY)
				maskaddr.sin_addr.s_addr = 0;
			if (!errflg)
				restrict(RESTRICT_FLAGS, &peeraddr, &maskaddr,
				    (int)peerkey, peerversion);
			break;

		case CONFIG_BDELAY:
			if (ntokens >= 2) {
				l_fp tmp;
				u_fp utmp;

				if (!atolfp(tokens[1], &tmp)) {
					tprintf(
				       "broadcastdelay value %s undecodable\n",
					    tokens[1]);
				} else if (tmp.l_ui != 0) {
					tprintf(
				       "broadcastdelay value %s is unlikely\n",
					    tokens[1]);
				} else {
					utmp = LFPTOFP(&tmp);
					proto_config(PROTO_BROADDELAY, utmp);
				}
			}
			break;
		
#ifdef XNTP_AUTHENTICATE
		case CONFIG_TRUSTEDKEY:
			for (i = 1; i < ntokens; i++) {
				u_long tkey;

				tkey = (u_long) atoi(tokens[i]);
				if (tkey == 0) {
					tprintf(
					    "trusted key %s unlikely\n",
					    tokens[i]);
				} else {
					authtrust(tkey, 1);
				}
			}
			break;

		case CONFIG_REQUESTKEY:
			if (ntokens >= 2) {
				u_long rkey;

				if (!atouint(tokens[1], &rkey)) {
					tprintf(
					  "%s is undecodeable as request key\n",
					    tokens[1]);
				} else if (rkey == 0) {
					tprintf(
					    "%s makes a poor request keyid\n",
					    tokens[1]);
				} else {
#ifdef DEBUG
					if (debug > 3)
						tprintf(
					"set info_auth_key to %lu\n", rkey);
#endif
					info_auth_keyid = rkey;
				}
			}
			break;
		
		case CONFIG_CONTROLKEY:
			if (ntokens >= 2) {
				u_long ckey;
				extern u_long ctl_auth_keyid;

				ckey = (u_long)atoi(tokens[1]);
				if (ckey == 0) {
					tprintf(
					    "%s makes a poor control keyid\n",
					    tokens[1]);
				} else {
					ctl_auth_keyid = ckey;
				}
			}
			break;
#endif  /* XNTP_AUTHENTICATE */
		
#ifdef XNTP_MODE6
		case CONFIG_TRAP:
			if (ntokens < 2) {
				tprintf(
				"no address for trap command, line ignored\n");
				  break;
			}
			if (!getnetnum(tokens[1], &peeraddr, 1))
				break;
			
			/*
			 * Use peerversion for port number.  Barf.
			 */
			errflg = 0;
			peerversion = 0;
			localaddr = 0;
			for (i = 2; i < ntokens-1; i++)
				switch (matchkey(tokens[i], trap_keywords)) {
				case CONF_TRAP_PORT:
					if (i >= ntokens-1) {
						tprintf(
					    "trap port requires an argument\n");
						errflg = 1;
						break;
					}
					peerversion = atoi(tokens[++i]);
					if (peerversion <= 0
					    || peerversion > 32767) {
						tprintf(
				      "invalid port number %s, trap ignored\n",
						    tokens[i]);
						errflg = 1;
					}
					break;

				case CONF_TRAP_INTERFACE:
					if (i >= ntokens-1) {
						tprintf(
				       "trap interface requires an argument\n");
						errflg = 1;
						break;
					}

					if (!getnetnum(tokens[++i],
					    &maskaddr, 1)) {
						errflg = 1;
						break;
					}

					localaddr = findinterface(&maskaddr);
					if (localaddr == NULL) {
						tprintf(
				      "can't find interface with address %s\n",
						    ntoa(&maskaddr));
						errflg = 1;
					}
					break;

				case CONFIG_UNKNOWN:
					errflg++;
					break;
				}

			if (!errflg) {
				extern struct interface *any_interface;

				if (peerversion != 0)
					peeraddr.sin_port = htons(peerversion);
				else
					peeraddr.sin_port = htons(TRAPPORT);
				if (localaddr == NULL)
					localaddr = any_interface;
				if (!ctlsettrap(&peeraddr, localaddr, 0,
				    NTP_VERSION))
					tprintf(
					"can't set trap for %s, no resources\n",
					    ntoa(&peeraddr));
			}
			break;
#endif  /* XNTP_MODE6 */
		
#ifdef REFCLOCK
		case CONFIG_FUDGE:
			if (ntokens < 2) {
				tprintf(
				"no address for fudge command, line ignored\n");
				  break;
			}
			if (!getnetnum(tokens[1], &peeraddr, 1))
				break;

			if (!ISREFCLOCKADR(&peeraddr)) {
				tprintf(
	"%s is inappropriate address for the fudge command, line ignored\n",
				    ntoa(&peeraddr));
				break;
			}

			bzero((char *)&clock, sizeof clock);
			errflg = 0;
			for (i = 2; i < ntokens-1; i++) {
				switch (c = matchkey(tokens[i],
				    fudge_keywords)) {
				case CONF_FDG_TIME1:
					if (!atolfp(tokens[++i],
					    &clock.fudgetime1)) {
						tprintf(
					"fudge %s time1 value in error\n",
						    ntoa(&peeraddr));
						errflg = i;
						break;
					}
					clock.haveflags |= CLK_HAVETIME1;
					break;

				case CONF_FDG_TIME2:
					if (!atolfp(tokens[++i],
					    &clock.fudgetime2)) {
						tprintf(
					"fudge %s time2 value in error\n",
						    ntoa(&peeraddr));
						errflg = i;
						break;
					}
					clock.haveflags |= CLK_HAVETIME2;
					break;

				case CONF_FDG_VALUE1:
					if (!atoint(tokens[++i],
					    &clock.fudgeval1)) {
						tprintf(
					"fudge %s value1 value in error\n",
						    ntoa(&peeraddr));
						errflg = i;
						break;
					}
					clock.haveflags |= CLK_HAVEVAL1;
					break;

				case CONF_FDG_VALUE2:
					if (!atoint(tokens[++i],
					    &clock.fudgeval2)) {
						tprintf(
					"fudge %s value2 value in error\n",
						    ntoa(&peeraddr));
						errflg = i;
						break;
					}
					clock.haveflags |= CLK_HAVEVAL2;
					break;

				case CONF_FDG_FLAG1:
				case CONF_FDG_FLAG2:
				case CONF_FDG_FLAG3:
				case CONF_FDG_FLAG4:
					if (!atoint(tokens[++i], &peerkey)
					    || peerkey > 1) {
						tprintf(
					"fudge %s flag value in error\n",
						    ntoa(&peeraddr));
						errflg = i;
						break;
					}
					switch(c) {
					case CONF_FDG_FLAG1:
						c = CLK_FLAG1;
						clock.haveflags|=CLK_HAVEFLAG1;
						break;
					case CONF_FDG_FLAG2:
						c = CLK_FLAG2;
						clock.haveflags|=CLK_HAVEFLAG2;
						break;
					case CONF_FDG_FLAG3:
						c = CLK_FLAG3;
						clock.haveflags|=CLK_HAVEFLAG3;
						break;
					case CONF_FDG_FLAG4:
						c = CLK_FLAG4;
						clock.haveflags|=CLK_HAVEFLAG4;
						break;
					}
					if (peerkey == 0)
						clock.flags &= ~c;
					else
						clock.flags |= c;
					break;

				case CONFIG_UNKNOWN:
					errflg = -1;
					break;
				}
			}

			/*
			 * If reference clock support isn't defined the
			 * fudge line will still be accepted and syntax
			 * checked, but will essentially do nothing.
			 */
			if (!errflg) {
				extern void refclock_control();

				refclock_control(&peeraddr, &clock,
				    (struct refclockstat *)0);
			}
			break;
#endif /* REFCLOCK */

		case CONFIG_MAXSKEW:
			if (ntokens >= 2) {
				l_fp tmp;
				u_fp utmp;

				if (!atolfp(tokens[1], &tmp)) {
					tprintf(
					 "maxskew value %s undecodable\n",
					    tokens[1]);
				} else if (tmp.l_ui != 0) {
					tprintf(
					  "maxskew value %s is unlikely\n",
					    tokens[1]);
				} else {
					utmp = LFPTOFP(&tmp);
					proto_config(PROTO_MAXSKEW, utmp);
				}
			}
			break;
		
		case CONFIG_SELECT:
			if (ntokens >= 2) {
				i = atoi(tokens[1]);
				if (i < SELECT_1 || i > SELECT_5)
					tprintf(
			      "invalid selection algorithm %s, line ignored\n",
					    tokens[1]);
				else
					proto_config(PROTO_SELECT, (long)i);
			}
			break;

		case CONFIG_DEBUG:
			if (ntokens == 1)
			  {
			    tprintf ("Debugging level is %d\n", debug);
			  }
			else
			  {
			    if (isdigit(*tokens[1]))  {
				    debug = atoi (tokens[1]);
				    if (debug > 255)
					    debug = 255;
			    } else {
				    switch (matchkey(tokens[1], debug_keywords))
				    {
				    case CONF_DBG_OFF:
					    debug = 0;
					    break;

				    case CONF_DBG_INC:
					    if (debug < 255)
						    debug++;
					    break;

				    case CONF_DBG_DEC:
					    if (debug > 0)
						    debug--;
					    break;

				    default:
				    tprintf("Should be `debug off|inc|dec'\n");
					    break;
				    }
			    }
		          }
			break;
		}
	}
}


/*
 * matchkey - match a keyword to a list
 */
static int
matchkey(word, keys)
	register char *word;
	register struct keyword *keys;
{
	for (;;) {
		if (keys->keytype == CONFIG_UNKNOWN) {
			tprintf(
			    "configure: keyword \"%s\" unknown, line ignored\n",
			    word);
			return CONFIG_UNKNOWN;
		}
		if (STREQ(word, keys->text))
			return keys->keytype;
		keys++;
	}
}


/*
 * getnetnum - return a net number (this is crude, but careful)
 */
static int
getnetnum(num, addr, complain)
	char *num;
	struct sockaddr_in *addr;
	int complain;
{
	u_long netnum;
	extern int32 resolve();
#ifdef DEBUG
	extern char *ntoa();
#endif

	netnum = resolve (num);
	if (!netnum)  {
		if (complain)
			tprintf(
		    "configure: \"%s\" not valid host number, line ignored\n",
			    num);
		return 0;
	}

	/*
	 * make up socket address.  Clear it out for neatness.
	 */
	bzero((char *)addr, sizeof(struct sockaddr_in));
	addr->sin_family = AF_INET;
	addr->sin_port = htons(NTP_PORT);
	addr->sin_addr.s_addr = htonl(netnum);
#ifdef DEBUG
	if (debug > 1)
		printf("getnetnum given %s, got %s (%lx)\n",
		    num, ntoa(addr), netnum);
#endif
	return 1;
}
